# Print a man page from the help texts.
#
# Copyright (C) 2011 Red Hat Inc.
# Author(s): Jesse Keating <jkeating@redhat.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.


import os
import sys
import datetime


# We could substitute the "" in .TH with the rpkg version if we knew it
man_header = """\
.\\" man page for rpkg
.TH rpkg 1 "%(today)s" "" "rpm\\-packager"
.SH "NAME"
rpkg \\- RPM packaging utility
.SH "SYNOPSIS"
.B "rpkg"
[
.I global_options
]
.I "command"
[
.I command_options
]
[
.I command_arguments
]
.br
.B "rpkg"
.B "help"
.br
.B "rpkg"
.I "command"
.B "\\-\\-help"
.SH "DESCRIPTION"
.B "rpkg"
is a script to manage spec-enriched Git repositories - meaning any repositories which contain
an rpm spec file. It is designed to work with expanded application sources as well as with standard DistGit
tarballs and patches. Note that you should trust the .spec files you work with because many operations
(like `rpkg srpm` or `rpkg lint`) involve parsing the spec file, which brings along evaluation
of any shell or lua scriplets present in it.
"""

man_footer = """\
.SH "CONFIGURATION"

rpkg reads configuration options from the following files in this order:

    /etc/rpkg.conf, ~/.config/rpkg.conf, <gitroot>/rpkg.conf, ./rpkg.conf

Configuration found in ./rpkg.conf has the highest priority. Configuration
in /etc/rpkg.conf has the lowest priority. <gitroot> placeholder above means
the top level of the Git repository we are working in. If we are in a plain
directory and all our parent directories are also plain, reading of <gitroot>/rpkg.conf
is simply skipped.

.SS "EXAMPLE CONFIGURATION"

    [rpkg]
    preprocess_spec = True

    # auto-packing is deprecated:
    auto_pack = False

    base_output_path = /tmp/rpkg

    # permanently sets --module-name option:
    # module_name = @group/project

    [git]
    lookaside = https://src.fedoraproject.org/repo/pkgs/%(ns1)s/%(pkg)s/%(filename)s/%(hashtype)s/%(hash)s/%(filename)s
    lookaside_cgi = https://src.fedoraproject.org/repo/pkgs/upload.cgi
    gitbaseurl = ssh://%(user)s@pkgs.fedoraproject.org/%(module)s
    anongiturl = https://src.fedoraproject.org/%(module)s

.SH "EXAMPLES"

Run %prep phase of an rpm build:

    $ rpkg prep
    git_pack: Wrote: /tmp/rpkg/rpkg-util-1-D4LQfm/rpkg-util-1.0.git.201.846d9fa.tar.gz
    Wrote: /tmp/rpkg/rpkg-util-1-D4LQfm/rpkg-util.spec
    Executing(%prep): /bin/sh -e /var/tmp/rpm-tmp.h0uU7f
    + umask 022
    + cd /tmp/rpkg/rpkg-util-1-D4LQfm
    + cd /tmp/rpkg/rpkg-util-1-D4LQfm
    + rm -rf rpkg-util-1.0.git.201.846d9fa
    + /usr/bin/gzip -dc /tmp/rpkg/rpkg-util-1-D4LQfm/rpkg-util-1.0.git.201.846d9fa.tar.gz
    + /usr/bin/tar -xof -
    + STATUS=0
    + '[' 0 -ne 0 ']'
    + cd rpkg-util-1.0.git.201.846d9fa
    + /usr/bin/chmod -Rf a+rX,u+w,g-w,o-w .
    + exit 0

For the following command, you need to have a project in COPR created. See quick tutorial at

https://docs.pagure.org/copr.copr/user_documentation.html#quick-start

on how to do it.

    $ rpkg build
    git_pack: Wrote: /tmp/rpkg/rpkg-util-2-kxBMs6/rpkg-util-1.0.git.201.846d9fa.tar.gz
    Wrote: /tmp/rpkg/rpkg-util-2-kxBMs6/rpkg-util.spec
    Wrote: /tmp/rpkg/rpkg-util-2-kxBMs6/rpkg-util-1.0.git.201.846d9fa-1.fc27.src.rpm
    Uploading package /tmp/rpkg/rpkg-util-2-kxBMs6/rpkg-util-1.0.git.201.846d9fa-1.fc27.src.rpm
    100% |################################| 114kB 329kB/s eta 0:00:00
    Build was added to rpkg-util:
      https://copr.fedorainfracloud.org/coprs/build/736534/
    Created builds: 736534
    Watching build(s): (this may be safely interrupted)
      20:42:20 Build 736534: importing
    ...

Here you can see a simple operation of generation spec file from a template:

    $ rpkg spec
    git_pack: Wrote: /tmp/rpkg/rpkg-util.spec

.SH "SPEC TEMPLATES"

Apart from being useful as a means to handle unpacked as well as packed content,
they also make it possible to generate certain parts of your spec file automatically
from a Git repository where your spec file is located. You need to have "spec_preprocess"
configuration option set to True in rpkg config to use this functionality.

.SS "BASIC EXAMPLE"

    Name:       {{{ git_name }}}
    Version:    {{{ git_version }}}
    Release:    1%{?dist}
    Summary:    This is a test package.

    License:    GPLv2+
    URL:        https://someurl.org
    VCS:        {{{ git_vcs }}}

    Source:     {{{ git_pack }}}

    %description
    This is a test package.

    %prep
    {{{ git_setup_macro }}}

    %changelog
    {{{ git_changelog }}}

Having a spec template like this in your project (instead of just a plain spec file)
allows you to build source and rpm packages automatically following git history of your project.

.SS "MACRO REFERENCE"

.TP

BASE GIT MACROS

Without any parameters, these macros work on repository level and use globally available repository information.
Output of every git macro call is cached and other macros are made to take those cached values into account by default.
You can see that a macro uses a cached value as a default input parameter value by looking at its function header definition and searching for the `cached` keyword.

E.g.

    git_version [name="$(cached git_name)"] [lead=0] [follow=]

Default value for name in this definition is derived from the cached output of the
previously invoked `git_name` macro. If `git_name` macro has not been invoked
or if its output was an empty string, the default value for the `name=` parameter
above will be also an empty string.

Note that $(cached_git_name_version) in the function header definitions below
is an alias for "$(cached git_name)-$(cached git_version)".

.B git_name [name=] [prepend=] [append=]

Without any parameters, lookup remote URL for the currently active
branch and output its basename with .git suffix stripped. If the
remote URL could not be determined, use "origin" remote URL
instead.

    Optional arguments:
        name     the base name to be output
        prepend  string to be prepended to the base name
        append   string to be appended to the base name

.B git_version [name="$(cached git_name)"] [lead=0] [follow=]

Output version of a subpackage given by ${name}. The version
is a string constructed of the following parts:

    "${lead}.${follow}${commit_count_appendix}${wtree_appendix}"

${lead} is a mandatory argument of the macro. By default it is set
to zero.

${follow} (if not specified) is obtained from the follow part of
the latest tag made for the subpackage given by ${name}.

${commit_count_appendix} is generated only if the current commit
is not tagged for the given subpackage and it is composed of the
following parts

    .git.<commit_count>.<latest_commit_short_hash>

Commit count is a number of commits from the latest tag on the
given subpackage.

${wtree_appendix} is generated only if the currently checked
out working tree is dirty. It is composed of the following parts:

    .wtree.<encoded_latest_file_status_change>

    Optional arguments:
        name     name of a subpackage
        lead     major version substring
        follow   minor version substring, dynamically generated if not specified

    Environment variables:
        VERSION_BUMP: if set to a non-empty value, the following procedure is applied
        to get the version string:

        If $follow is specified by user explicitly, return $lead.$follow.

        Otherwise, extract $follow from the latest tag as the last version component (after the last dot)
        and return $lead.$follow+1. If $follow is not a number, throw an error (manual setting of follow
        is needed).

        rpkg sets the VERSION_BUMP variable into the preprocessing environment during
        execution of `rpkg tag` subcommand. Otherwise, it is unset.

.B git_vcs [subtree=]

Output pseudo repository URL pointing the currently checked out tree.

Example output: git+ssh://git@pagure.io/test-project.git#dadef2a8b9554e94797a7336261192e02d5d9351:

    Optional arguments:
        subtree     path to a subtree to generate the URL for

.B git_pack [path="$(git rev-parse --show-toplevel)"] [dir_name="$(cached_git_name_version)"] [source_name="$(cached_git_name_version).tar.gz"]

Pack the whole working tree content into a gzipped source tarball and output $source_name, which is a filename of the created tarball.
Uses `tar caf --exclude-vcs --exclude-vcs-ignores` to do the job.

    Optional arguments:
        path         path to a specific subdirectory to be packed
        dir_name     top-level directory name in the created tarball
        source_name  filename of the created source tarball

    Enviroment variables:
        OUTDIR: if empty, git_pack will not generate any sources, only outputs the source filename

        rpkg sets OUTDIR, only if the tarball is actually needed to be generated. It does not set it
        when `rpkg spec` is called (without any additional arguments).

.B git_archive [path="$(git rev-parse --show-toplevel)"] [dir_name="$(cached_git_name_version)"] [source_name="$(cached_git_name_version).tar.gz"]

Pack the whole working tree content into a gzipped source tarball and output $source_name, which is a filename of the created tarball.
Uses `git archive` to do the job. Note that this macro will refuse to generate a tarball if your working tree is dirty.

    Optional arguments:
        path         path to a specific subdirectory to be packed
        dir_name     top-level directory name in the created tarball
        source_name  filename of the created source tarball

    Enviroment variables:
        OUTDIR: if empty, git_pack will not generate any sources, only outputs the source filename

        rpkg sets OUTDIR, only if the tarball is actually needed to be generated. It does not set it
        when `rpkg spec` is called (without any additional arguments).

.B git_setup_macro [dir_name="$(cached_git_name_version)"]

Output %setup rpm macro for the given top-level source-tarball directory.

    Optional arguments:
        dir_name     name of the top-level directory in the source-tarball

.B git_changelog [name="$(cached git_name)"] [since_tag=] [until_tag=] [header_locale=POSIX] [header_date_format="%a %b %d %Y"] [body_wrap=80]

Output rpm spec changelog as generated from tag messages for the subpackage given by ${name}.

    Optional arguments:
        name                name of the subpackage to generate the changelog for
        since_tag           start the changelog records with this tag
        until_tag           end the changelog records with this tag
        header_locale       locale to be used when changelog record date is being generated
        header_date_format  date format to be used in the changelog record headers
        body_wrap           maximum allowed line length for changelog body

When `rpkg tag` is called, it allows you to specify a tag message. This is the message
that will be used as a changelog record body.

.TP

GIT_DIR MACROS

These macros operate on a directory level. In particular, on a directory where the input spec file template is located. They are basically
aliases to the GIT BASE MACROS with some input values set specifically for the directory-level operation. The input spec template location is
given by INPUT_PATH environment variable set by rpkg. You can influence it from the command line by explicitly specifying --spec <template_path>
argument for rpkg subcommands that support it.

Below you can see to what the GIT DIR macros expand. In some cases, the macro is a direct alias to the underlying git base macro.

.B git_dir_name

    git_name append="-$(git -C "$(dirname "$INPUT_PATH")" rev-parse --show-prefix | name_convert)"

    Outputs repository origin URL basename appended with the path (after slash to dash substitution) from the repository
    root to the spec file's directory.

.B git_dir_version

    git_version

.B git_dir_vcs

    git_vcs subtree="$(git -C "$(dirname "$INPUT_PATH")" rev-parse --show-prefix)"

    Outputs VCS pseudo URL pointing to the spec file's subdirectory.

.B git_dir_pack

    git_pack path="$(dirname "$INPUT_PATH")"

    Packs spec file's subdirectory by using `tar`.

.B git_dir_archive

    git_archive path="$(dirname "$INPUT_PATH")"

    Packs spec file's subdirectory by using `git archive`.

.B git_dir_setup_macro

    git_setup_macro

.B git_dir_changelog

    git_changelog

.TP

GIT_CWD MACROS

These macros operate on a directory level. In particular, on a rpkg's working directory. Again, they are
aliases to the GIT BASE MACROS with some input values set specifically for the directory-level operation.

You can influence the rpkg's working dir from the command line by explicitly specifying its --path argument.

Below you can see to what the GIT_CWD macros expand. In some cases, the macro is a direct alias for the underlying git base macro.

.B git_cwd_name

    git_name append="-$(git -C "$(pwd)" rev-parse --show-prefix | name_convert)"

    Outputs repository origin URL basename appended with the rpkg's working directory path (after slash to dash substitution).

.B git_cwd_version

    git_version

.B git_cwd_vcs

    git_vcs subtree="$(git -C "$(pwd)" rev-parse --show-prefix)"

    Outputs VCS pseudo URL pointing to the rpkg's working directory.

.B git_cwd_pack

    git_pack path="$(pwd)"

    Packs rpkg's working directory by using `tar`.

.B git_cwd_archive

    git_archive path="$(pwd)"

    Packs rpkg's working directory by using `git archive`.

.B git_cwd_setup_macro

    git_setup_macro

.B git_cwd_changelog

    git_changelog

.SH USER-DEFINED MACROS

You can define your own macros (basically just bash functions) by putting them
into rpkg.macros file in rpkg's working directory. You can even override the
default macros provided by rpkg.

.SH AUTO-PACKING (DEPRECATED)

Auto-packing works only if auto_pack configuration option is set to True and
only if the content you are working with is not packed (see `rpkg is-packed`).

When invoking `rpkg srpm` on an unpacked content, rpkg will download external sources
(if any) and then it will generate a tarball from the whole content of the working
directory named according to Source0 definition present in the associated .spec file.
This tarball and the .spec file are then passed to rpmbuild to generate the resulting
srpm.

.SH "PACKED VS UNPACKED"

NOTE: This article relates to deprecated auto_pack feature. You can safely ignore it.

While it is quite intuitive what is packed content (.spec + tarballs + patches)
and what is unpacked content (.spec + original application source files), it
might be useful to know how exactly rpkg differentiates between these two.
You can go through the following examples to get overview how this tool exactly
works.

.SS PACKED CONTENT:

    $ cd source0-present-example
    $ grep -E '(Source|Patch)' testpkg.spec
    Source0: foo.tar.gz
    $ ls .
    foo.tar.gz testpkg.spec
    $ rpkg make-source
    Could not execute make_source: Not an unpacked content.
    $ rpkg srpm
    Failed to get module name from Git url or pushurl.
    Wrote: source0-present-example/testpkg-1-1.fc25.src.rpm

    The error about module name is caused by running `rpkg`
    on a plain directory content and not a Git repo. In this
    case module name is read out from the spec file. The
    error about not being able to make source is expected for
    packed content (tarballs to be put into srpm are expected
    to be present already).

    $ cd only-ignored-files-example
    $ grep -E '(Source|Patch)' testpkg.spec
    Source0: https://example.org/foo.tar.gz
    $ ls .
    README.md testpkg.spec
    $ rpkg make-source
    Could not execute make_source: Not an unpacked content.
    $ echo '%_disable_source_fetch 0' >> ~/.rpmmacros
    $ rpkg srpm
    Failed to get module name from Git url or pushurl
    warning: Downloading https://example.org/foo.tar.gz to only-ignored-files-example/foo.tar.gz
    Wrote: only-ignored-files-example/testpkg-1-1.fc25.src.rpm

    In this example, sources are downloaded from network when
    srpm is being built. The %_disable_source_fetch rpm macro
    must be set to 0 and the tarball url must be valid for this
    to work. The content is recognized as packed because there
    are only ignored files in the directory (.spec and readmes).

.SS UNPACKED CONTENT:

    $ cd cpp-source-file-present-example
    $ git init .
    $ git remote add origin https://github.com/testpkg.git
    $ grep -E '(Source|Patch)' testpkg.spec
    Source0: foo.tar.gz
    $ ls .
    main.cpp testpkg.spec
    $ rpkg make-source
    Wrote: cpp-source-file-present-example/foo.tar.gz
    $ rpkg srpm
    Wrote: cpp-source-file-present-example/testpkg-1-1.fc25.src.rpm

    foo.tar.gz (the only Source referenced from the .spec file) is
    not present and there is unignored main.cpp file that makes the
    content recognized as unpacked. When `rpkg make-source` is invoked,
    foo.tar.gz is created and will contain the main.cpp file (as
    well as the .spec metadata file but that is just because the
    whole content directory is packed). Note that the error about
    failing to get module name from Git url disappeared because
    we have run `git init .` and `git remote add ...`.

.SH "SEE ALSO"
.UR "https://pagure.io/rpkg-util/"
.BR "https://pagure.io/rpkg-util/"
"""


class ManFormatter(object):
    def __init__(self, man):
        self.man = man

    def write(self, data):
        for line in data.split('\n'):
            self.man.write('  %s\n' % line)


def strip_usage(s):
    """Strip "usage: " string from beginning of string if present"""
    if s.startswith('usage: '):
        return s.replace('usage: ', '', 1)
    else:
        return s


def man_constants():
    """Global constants for man file templates"""
    today = datetime.date.today()
    today_manstr = today.strftime(r'%Y\-%m\-%d')
    return {'today': today_manstr}


def generate(parser, subparsers):
    """\
    Generate the man page on stdout

    Given the argparse based parser and subparsers arguments, generate
    the corresponding man page and write it to stdout.
    """

    # Not nice, but works: Redirect any print statement output to
    # stderr to avoid clobbering the man page output on stdout.
    man_file = sys.stdout
    sys.stdout = sys.stderr

    mf = ManFormatter(man_file)

    choices = subparsers.choices
    k = sorted(choices.keys())

    man_file.write(man_header % man_constants())

    helptext = parser.format_help()
    helptext = strip_usage(helptext)
    helptextsplit = helptext.split('\n')
    helptextsplit = [line for line in helptextsplit
                     if not line.startswith('  -h, --help')]

    man_file.write('.SS "%s"\n' % ("Global Options",))

    outflag = False
    for line in helptextsplit:
        if line == "optional arguments:":
            outflag = True
        elif line == "":
            outflag = False
        elif outflag:
            man_file.write("%s\n" % line)

    help_texts = {}
    for pa in subparsers._choices_actions:
        help_texts[pa.dest] = getattr(pa, 'help', None)

    man_file.write('.SH "COMMAND OVERVIEW"\n')

    for command in k:
        cmdparser = choices[command]
        usage = cmdparser.format_usage()
        usage = strip_usage(usage)
        usage = ''.join(usage.split('\n'))
        usage = ' '.join(usage.split())
        if help_texts[command]:
            man_file.write('.TP\n.B "%s"\n%s\n' % (usage, help_texts[command]))
        else:
            man_file.write('.TP\n.B "%s"\n' % (usage))

    man_file.write('.SH "COMMAND REFERENCE"\n')
    for command in k:
        cmdparser = choices[command]

        man_file.write('.SS "%s"\n' % cmdparser.prog)

        help = help_texts[command]
        if help and not cmdparser.description:
            if not help.endswith('.'):
                help = "%s." % help
            cmdparser.description = help

        h = cmdparser.format_help()
        mf.write(h)

    man_file.write(man_footer)


if __name__ == '__main__':
    module_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    sys.path.insert(0, module_path)

    from rpkglib.cli import rpkgClient
    client = rpkgClient(name='rpkg', config=None)
    client.do_imports('rpkglib')

    generate(client.parser, client.subparsers)
